## [-12.8, 12.8]
# RESOLUTION = 0.2
# ENV_SIZE = 128
RESOLUTION = 0.1
ENV_SIZE = 256
# RESOLUTION = 0.05
# ENV_SIZE = 512

PASSABLE_COLOR = 0
MAP_OBS_VAL = 100
MAP_PASS_VAL = 0
MAP_UNK_VAL = -1

SPAWN_EXTENSION = 0.3